package at.jku.ssw.list.test;

import static org.junit.Assert.assertEquals;

import java.util.ArrayList;
import java.util.NoSuchElementException;

import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

import at.jku.ssw.Iterator;
import at.jku.ssw.list.List;

public abstract class ListTest {

	protected abstract List createList();

	protected List createList(Object... objects) {
		List list = createList();
		addValues(list, objects);
		return list;
	}

	private void addValues(List list, Object[] objects) {
		for (int i = objects.length - 1; i >= 0; i--) {
			list.insertFirst(objects[i]);
		}
	}

	protected void assertContains(List list, Object... objects) {
		Iterator it = list.iterator();
		int i = 0;
		while (it.hasNext()) {
			Assert.assertEquals(objects[i++], it.next());
		}
		Assert.assertEquals(objects.length, i);
	}

	@Rule
	public ExpectedException thrown = ExpectedException.none();

	@Test
	public void testInsertFirst() {
		List list = createList();
		list.insertFirst(3);
		list.insertFirst(2);
		list.insertFirst(1);
		assertContains(list, 1, 2, 3);
	}

	@Test
	public void testInsertBegin() {
		List list = createList(1, 2, 3);
		list.insert(0, 99);
		list.insert(0, 101);
		assertContains(list, 101, 99, 1, 2, 3);
	}

	@Test
	public void testInsertMiddle() {
		List list = createList(1, 2, 3, 5, 6);
		list.insert(3, 4);
		assertContains(list, 1, 2, 3, 4, 5, 6);
	}

	@Test
	public void testInsertWithLast() {
		List list = createList(1, 2, 3);
		list.insert(3, 4);
		assertContains(list, 1, 2, 3, 4);
	}

	@Test
	public void testInsertTooLarge() {
		List list = createList(1, 2, 3);
		thrown.expect(IndexOutOfBoundsException.class);
		list.insert(100, 1);
	}

	@Test
	public void testInsertNegative() {
		List list = createList(1, 2, 3);
		thrown.expect(IndexOutOfBoundsException.class);
		list.insert(-1, 1);
	}

	@Test
	public void testGetEmptyList() {
		List list = createList();
		thrown.expect(IndexOutOfBoundsException.class);
		list.get(0);
	}

	@Test
	public void testGetOverflow() {
		List list = createList(1, 2, 3);
		thrown.expect(IndexOutOfBoundsException.class);
		list.get(list.size());
	}

	@Test
	public void testGetUnderflow() {
		List list = createList(1, 2, 3);
		thrown.expect(IndexOutOfBoundsException.class);
		list.get(-1);
	}

	@Test
	public void testGetList() {
		List list = createList(1, 2, 3);

		assertEquals(1, list.get(0));
		assertEquals(2, list.get(1));
		assertEquals(3, list.get(2));

	}

	@Test
	public void testRemoveEmpty() {
		List list = createList();
		thrown.expect(IndexOutOfBoundsException.class);
		list.remove(0);
	}

	@Test
	public void testRemoveUnderflow() {
		List list = createList(5, 6, 7);
		thrown.expect(IndexOutOfBoundsException.class);
		list.remove(-1);
	}

	@Test
	public void testRemoveOverflow() {
		List list = createList(5, 6, 7);
		thrown.expect(IndexOutOfBoundsException.class);
		list.remove(list.size());
	}

	@Test
	public void testRemove() {
		List list = createList(1, 2, 3, 4);
		list.remove(1);
		list.remove(1);
		assertContains(list, 1, 4);
	}

	@Test
	public void testRemoveFirst() {
		List list = createList(1, 2, 3, 4);
		list.remove(0);
		assertContains(list, 2, 3, 4);
	}

	@Test
	public void testRemoveTail() {
		List list = createList(1, 2, 3, 4);
		list.remove(3);
		assertContains(list, 1, 2, 3);
	}

	@Test
	public void testRemoveToEmptyList() {
		List list = createList(1);
		list.remove(0);
		assertEquals(0, list.size());
	}

	@Test
	public void testIndexOfEmpty() {
		List list = createList();
		assertEquals(-1, list.indexOf(1));
	}

	@Test
	public void testIndexOf() {
		List list = createList(1, 2, 3, 4, 5);
		assertEquals(1, list.indexOf(2));
	}

	@Test
	public void lastIndexOfEmpty() {
		List list = createList();
		assertEquals(-1, list.indexOf(1));
	}

	@Test
	public void lastIndexOf() {
		List list = createList(1, 2, 3, 4, 5, 6, 7);
		assertEquals(4, list.lastIndexOf(5));
	}

	@Test
	public void lastIndexOfSameElements() {
		List list = createList(1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2);
		assertEquals(6, list.lastIndexOf(1));
	}

	@Test
	public void lastIndexOfNotFound() {
		List list = createList(1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2);
		assertEquals(-1, list.lastIndexOf(77));
	}

	@Test
	public void size() {
		List list = createList(1, 2, 3);
		assertEquals(3, list.size());
	}

	@Test
	public void sizeEmpty() {
		List list = createList();
		assertEquals(0, list.size());
	}

	@Test
	public void sizeOneElement() {
		List list = createList(1);
		assertEquals(1, list.size());
	}

	@Test
	public void iterator() {
		List list = createList(1, 2, 3);
		Iterator it = list.iterator();
		ArrayList<Object> orig = new ArrayList<>();
		orig.add(1);
		orig.add(2);
		orig.add(3);
		ArrayList<Object> l = new ArrayList<>();
		while (it.hasNext()) {
			l.add(it.next());
		}
		assertEquals(orig, l);
	}

	@Test
	public void iteratorEmpty() {
		List list = createList();
		Iterator it = list.iterator();
		assertEquals(false, it.hasNext());
		thrown.expect(NoSuchElementException.class);
		it.next();
	}

	@Test
	public void testRemoveFirstEmpty() {
		List list = createList();
		thrown.expect(IndexOutOfBoundsException.class);
		list.removeFirst();
	}

}
